unit MainFormUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, XPMan, Math, ExtCtrls;

type
  TMainForm = class(TForm)
    EditFile: TEdit;
    ButtonFile: TButton;
    OpenDialog: TOpenDialog;
    XPManifest1: TXPManifest;
    ButtonProcess: TButton;
    RadioGroup: TRadioGroup;
    GroupBox1: TGroupBox;
    LabelDemension: TLabel;
    LabelOffset: TLabel;
    LabelCluster: TLabel;
    EditDemension: TEdit;
    UpDownDemension: TUpDown;
    EditOffset: TEdit;
    UpDownOffset: TUpDown;
    EditCluster: TEdit;
    UpDownCluster: TUpDown;
    StatusBar: TStatusBar;
    procedure ButtonFileClick(Sender: TObject);
    procedure ButtonProcessClick(Sender: TObject);
  private
    { Private 錾 }
  public
    { Public 錾 }
  end;

type
  TVector = array of Double;
  TVectorList = array of TVector;
  TDistanceFunction = function(VectorList: TVectorList; VectorIndexList1, VectorIndexList2: array of Integer): Double;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

function EuclidDistance(Vector1, Vector2: TVector): Double;
var
  Cnt: Integer;
  Demension: Integer;
begin
  Demension := Min(Length(Vector1), Length(Vector2));
  Result := 0;
  for Cnt := 0 to Demension - 1 do
    Result := Result + Power(Vector1[Cnt] - Vector2[Cnt], 2);
  Result := Sqrt(Result);
end;

// (ŒZ@)

function DistanceNearest(VectorList: TVectorList; VectorIndexList1, VectorIndexList2: array of Integer): Double;
var
  Cnt1, Cnt2: Integer;
begin
  Result := MaxDouble;
  for Cnt1 := 0 to Length(VectorIndexList1) - 1 do
    for Cnt2 := 0 to Length(VectorIndexList2) - 1 do
      Result := Min(Result, EuclidDistance(VectorList[Cnt1], VectorList[Cnt2]));
end;

// (ŒZ@)

function DistanceFurthest(VectorList: TVectorList; VectorIndexList1, VectorIndexList2: array of Integer): Double;
var
  Cnt1, Cnt2: Integer;
begin
  Result := 0;
  for Cnt1 := 0 to Length(VectorIndexList1) - 1 do
    for Cnt2 := 0 to Length(VectorIndexList2) - 1 do
      Result := Max(Result, EuclidDistance(VectorList[Cnt1], VectorList[Cnt2]));
end;

// (Qϖ@)

function DistanceAverage(VectorList: TVectorList; VectorIndexList1, VectorIndexList2: array of Integer): Double;
var
  Cnt1, Cnt2: Integer;
begin
  Result := 0;
  for Cnt1 := 0 to Length(VectorIndexList1) - 1 do
    for Cnt2 := 0 to Length(VectorIndexList2) - 1 do
      Result := Result + EuclidDistance(VectorList[Cnt1], VectorList[Cnt2]);
  Result := Result / (Length(VectorIndexList1) * Length(VectorIndexList2));
end;

// (EH[h@)

function DistanceWard(VectorList: TVectorList; VectorIndexList1, VectorIndexList2: array of Integer): Double;
  // NX^dS
  procedure CalcCenter(VectorList: TVectorList; var Vector: TVector);
  var
    Cnt1, Cnt2: Integer;
  begin
    for Cnt1 := 0 to Length(Vector) - 1 do
      Vector[Cnt1] := 0;
    for Cnt1 := 0 to Length(VectorList) - 1 do
      for Cnt2 := 0 to Length(Vector) - 1 do
        Vector[Cnt2] := Vector[Cnt2] + VectorList[Cnt1][Cnt2];
    for Cnt1 := 0 to Length(Vector) - 1 do
      Vector[Cnt1] := Vector[Cnt1] / Length(VectorList);
  end;
var
  Cnt, MinDemension, TmpOffset: Integer;
  TmpVectorList: TVectorList;
  VectorCenter: TVector;
begin
  Result := 0;
  MinDemension := MaxInt;
  // NX^
  SetLength(TmpVectorList, Length(VectorIndexList1) + Length(VectorIndexList2));
  for Cnt := 0 to Length(VectorIndexList1) - 1 do
  begin
    TmpVectorList[Cnt] := VectorList[VectorIndexList1[Cnt]];
    MinDemension := Min(MinDemension, Length(VectorList[VectorIndexList1[Cnt]]));
  end;
  TmpOffset := Length(VectorIndexList1);
  for Cnt := 0 to Length(VectorIndexList2) - 1 do
  begin
    TmpVectorList[Cnt + TmpOffset] := VectorList[VectorIndexList2[Cnt]];
    MinDemension := Min(MinDemension, Length(VectorList[VectorIndexList2[Cnt]]));
  end;
  SetLength(VectorCenter, MinDemension);
  CalcCenter(TmpVectorList, VectorCenter);
  for Cnt := 0 to Length(TmpVectorList) - 1 do
    Result := Result + Power(EuclidDistance(TmpVectorList[Cnt], VectorCenter), 2);
  // NX^1
  SetLength(TmpVectorList, Length(VectorIndexList1));
  for Cnt := 0 to Length(VectorIndexList1) - 1 do
    TmpVectorList[Cnt] := VectorList[VectorIndexList1[Cnt]];
  CalcCenter(TmpVectorList, VectorCenter);
  for Cnt := 0 to Length(TmpVectorList) - 1 do
    Result := Result - Power(EuclidDistance(TmpVectorList[Cnt], VectorCenter), 2);
  // NX^2
  SetLength(TmpVectorList, Length(VectorIndexList2));
  for Cnt := 0 to Length(VectorIndexList2) - 1 do
    TmpVectorList[Cnt] := VectorList[VectorIndexList2[Cnt]];
  CalcCenter(TmpVectorList, VectorCenter);
  for Cnt := 0 to Length(TmpVectorList) - 1 do
    Result := Result - Power(EuclidDistance(TmpVectorList[Cnt], VectorCenter), 2);
end;

procedure TMainForm.ButtonFileClick(Sender: TObject);
begin
  if OpenDialog.Execute then
    EditFile.Text := OpenDialog.FileName;
end;

procedure TMainForm.ButtonProcessClick(Sender: TObject);
  // t@Cϊ
  function MakeFileName(BaseName, Decoration: string): string;
  var
    FileExt, FileName: string;
  begin
    FileExt := ExtractFileExt(BaseName);
    FileName := ChangeFileExt(ExtractFileName(BaseName), '');
    Result := ExtractFilePath(BaseName) + FileName + Decoration + FileExt;
  end;
var
  InputCsv, OutputCsv, CommaLine: TStringList;
  CntX, CntY, CntY1, CntY2, CountCluster: Integer;
  Offset, Demension, Cluster: Integer;
  MinPos1, MinPos2, TmpOffset: Integer;
  MinValue, TmpValue: Double;
  VactorIndexList: array of array of Integer;
  VectorList: TVectorList;
  DistanceFunction: TDistanceFunction;
begin
  EditFile.Enabled := False;
  ButtonFile.Enabled := False;
  EditCluster.Enabled := False;
  UpDownCluster.Enabled := False;
  EditDemension.Enabled := False;
  UpDownDemension.Enabled := False;
  EditOffset.Enabled := False;
  UpDownOffset.Enabled := False;
  RadioGroup.Enabled := False;
  // 
  Demension := UpDownDemension.Position;
  Offset := UpDownOffset.Position;
  Cluster := UpDownCluster.Position;
  InputCsv := TStringList.Create;
  OutputCsv := TStringList.Create;
  CommaLine := TStringList.Create;
  case RadioGroup.ItemIndex of
    1: DistanceFunction := DistanceNearest;
    2: DistanceFunction := DistanceFurthest;
    3: DistanceFunction := DistanceAverage;
  else
    DistanceFunction := DistanceWard;
  end;
  try
    InputCsv.LoadFromFile(EditFile.Text);
    // sv̈Jbg+f[^
    SetLength(VectorList, InputCsv.Count);
    for CntY := 0 to Length(VectorList) - 1 do
    begin
      CommaLine.CommaText := InputCsv[CntY];
      SetLength(VectorList[CntY], Demension);
      for CntX := 0 to Demension - 1 do
        VectorList[CntY][CntX] := StrToFloat(CommaLine[CntX + Offset]);
    end;
    // NX^|C^
    SetLength(VactorIndexList, InputCsv.Count);
    for CntY := 0 to Length(VactorIndexList) - 1 do
    begin
      SetLength(VactorIndexList[CntY], 1);
      VactorIndexList[CntY][0] := CntY;
    end;
    // NX^OJn
    while Length(VactorIndexList) > Cluster do
    begin
      CountCluster := Length(VactorIndexList);
      StatusBar.SimpleText := 'Remain : ' + IntToStr(CountCluster) + ' clusters';
      Application.ProcessMessages;
      MinValue := MaxDouble;
      MinPos1 := 0;
      MinPos2 := 0;
      for CntY1 := 0 to CountCluster - 2 do
      begin
        for CntY2 := CntY1 + 1 to CountCluster - 1 do
        begin
          TmpValue := DistanceFunction(VectorList, VactorIndexList[CntY1], VactorIndexList[CntY2]);
          if MinValue > TmpValue then
          begin
            MinValue := TmpValue;
            MinPos1 := CntY1;
            MinPos2 := CntY2;
          end;
        end;
      end;
      // NX^
      TmpOffset := Length(VactorIndexList[MinPos1]);
      SetLength(VactorIndexList[MinPos1], Length(VactorIndexList[MinPos1]) + Length(VactorIndexList[MinPos2]));
      for CntX := 0 to Length(VactorIndexList[MinPos2]) - 1 do
        VactorIndexList[MinPos1][CntX + TmpOffset] := VactorIndexList[MinPos2][CntX];
      SetLength(VactorIndexList[MinPos2], 0);
      for CntY := MinPos2 + 1 to Length(VactorIndexList) - 1 do
        VactorIndexList[CntY - 1] := VactorIndexList[CntY];
      SetLength(VactorIndexList, Length(VactorIndexList) - 1);
    end;
    // ۑ
    for CntY := 0 to Length(VactorIndexList) - 1 do
    begin
      OutputCsv.Clear;
      for CntX := 0 to Length(VactorIndexList[CntY]) - 1 do
        OutputCsv.Add(InputCsv[VactorIndexList[CntY][CntX]]);
      OutputCsv.SaveToFile(MakeFileName(EditFile.Text, '-cluster' + IntToStr(CntY)));
    end;
  finally
    InputCsv.Free;
    OutputCsv.Free;
    CommaLine.Free;
    EditFile.Enabled := True;
    ButtonFile.Enabled := True;
    EditCluster.Enabled := True;
    UpDownCluster.Enabled := True;
    EditDemension.Enabled := True;
    UpDownDemension.Enabled := True;
    EditOffset.Enabled := True;
    UpDownOffset.Enabled := True;
    RadioGroup.Enabled := True;
  end;
end;

end.

